import gymnasium as gym
import numpy as np
from gymnasium import spaces


class PickHigh(gym.Env):
    metadata = {"render_modes": ["ansi"], "render_fps": 4}

    def __init__(self, render_mode=None):
        self.observation_space = spaces.Box(0, 9, (2,), dtype=np.int8)
        self.action_space = spaces.Discrete(2)

        assert render_mode is None or render_mode in self.metadata["render_modes"]
        self.render_mode = render_mode

    def _roll(self):
        left = self.np_random.integers(0, 10)
        right = self.np_random.integers(0, 10)
        self.draw = [left, right]

    def reset(self, seed=None, options=None):
        super().reset(seed=seed)
        self._roll()
        observation = np.array(self.draw, dtype=np.int8)
        info = {}
        return observation, info

    def step(self, action):
        left = self.draw[0]
        right = self.draw[1]
        card_player = left if action == 0 else right
        card_dealer = right if action == 0 else left

        terminated = bool(card_player != card_dealer)

        observation = np.array(self.draw, dtype=np.int8)
        reward = 1 if card_player > card_dealer else 0 if card_player == card_dealer else -1
        truncated = False
        info = {}

        if not terminated:
            self._roll()

        return observation, reward, terminated, truncated, info

    def render(self):
        return f"{self.draw[0]} {self.draw[1]}"



import gymnasium as gym
import numpy as np
from gymnasium import spaces
import random


class Shuffler(gym.Env):
    metadata = {"render_modes": ["ansi"], "render_fps": 1}

    def __init__(self):
        self.state = None
        self.observation_space = spaces.Box(
            low=0,
            high=5,
            shape=(6,),
            dtype=np.int8
        )
        self.action_space = spaces.Discrete(6)

    def _get_random_state(self):
        state = [i for i in range(6)]
        random.shuffle(state)
        return np.array(state, dtype=np.int8)

    def reset(self, seed=None, options=None):
        super().reset(seed=seed)
        self.state = self._get_random_state()
        return self.state, {}

    def step(self, action):
        space = np.where(self.state == 5)
        self.state[action], self.state[space] = self.state[space].item(), self.state[action].item()
        obs = self.state
        is_complete = np.array_equal(self.state, np.array([i for i in range(6)]))
        reward = 1 if is_complete else 0
        terminated = is_complete
        truncated = False
        return obs, reward, terminated, truncated, {}

    def render(self):
        return ' '.join(self.state)



import gymnasium as gym
from gymnasium import spaces


class TrafficLight(gym.Env):
    metadata = {"render_modes": ["ansi"], "render_fps": 4}

    def __init__(self):
        self.state = None
        self.observation_space = spaces.Discrete(3)
        self.action_space = spaces.Discrete(3)

    def _get_color(self, state):
        return ["green", "yellow", "red"][state]

    def reset(self, seed=None, options=None):
        super().reset(seed=seed)

        # Reset the state to green
        self.state = 0
        return self.state, {"color": self._get_color(self.state)}

    # state: 0: green 1: yellow 2: red
    # action: 0: drive 1: slow 2: stop
    def step(self, action):

        reward_structure = {
            0: {0: 1, 1: -1, 2: -1},
            1: {0: -1, 1: 1, 2: -1},
            2: {0: -1, 1: -1, 2: 1}
        }

        # calculate reward
        reward = reward_structure[self.state][action]
        terminated = self.state == 2
        truncated = reward == -1

        # identify the next state
        # generate a random chance to change the state
        roll = self.np_random.random()
        if not terminated and not truncated and roll > 0.7:
            self.state = (self.state + 1) % 3

        return self.state, reward, terminated, truncated, {"color": self._get_color(self.state)}

    def render(self):
        return self._get_color(self.state)

    def close(self):
        pass



import gymnasium as gym
from gymnasium import spaces
import math


class Cannon(gym.Env):
    metadata = {"render_modes": ["ansi"], "render_fps": 1}

    def __init__(self):
        self.shots = 10
        self.distance = 0
        self.observation_space = spaces.Box(10, 100, (1, 1))
        self.action_space = spaces.Box(1, 89)

    def reset(
            self,
            seed=None,
            options=None
    ):
        super().reset(seed=seed)
        self.shots = 10
        self._roll_distance()
        return self.distance, {"shots_left": self.shots}

    def _roll_distance(self):
        self.distance = self.np_random.random() * 90 + 10

    def step(self, action):
        self.shots -= 1
        fired_distance = self._calculate_projectile_distance(action)
        error = abs(fired_distance - self.distance)
        hit = error < 0.2

        terminated = self.shots == 0
        truncated = False

        if not terminated:
            self._roll_distance()

        obs = self.distance
        reward = 1 if hit else 0
        return obs, reward, terminated, truncated, {"shots_left": self.shots}

    def _calculate_projectile_distance(self, angle_degrees, speed=32):
        # Convert angle from degrees to radians
        angle_radians = math.radians(angle_degrees)

        # Gravity (m/s^2)
        g = 9.81

        # Calculate distance using the projectile motion formula
        distance = (speed ** 2) * math.sin(2 * angle_radians) / g

        return distance
